@@ -2,6 +2,8 @@
 
 /* 
  * Copyright (C) 2001 Havoc Pennington
+ * Copyright (C) 2002, 2003 Red Hat, Inc.
+ * Copyright (C) 2003 Rob Adams
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -88,6 +90,8 @@ find_next_cascade (MetaWindow *window,
   GList *sorted;
   int cascade_x, cascade_y;
   int x_threshold, y_threshold;
+  int window_width, window_height;
+  int cascade_stage;
   MetaRectangle work_area;  
   
   sorted = g_list_copy (windows);
@@ -98,36 +102,42 @@ find_next_cascade (MetaWindow *window,
    * new window after it. If a window is already nearly at that
    * position, we move on.
    */
-
+  
+  /* arbitrary-ish threshold, honors user attempts to
+   * manually cascade.
+   */
+#define CASCADE_FUZZ 15
+  if (fgeom)
+    {
+      x_threshold = MAX (fgeom->left_width, CASCADE_FUZZ);
+      y_threshold = MAX (fgeom->top_height, CASCADE_FUZZ);
+    }
+  else
+    {
+      x_threshold = CASCADE_FUZZ;
+      y_threshold = CASCADE_FUZZ;
+    }
+  
   /* Find furthest-SE origin of all workspaces.
    * cascade_x, cascade_y are the target position
    * of NW corner of window frame.
    */
-  /* FIXME this is bogus because we get the current xinerama
-   * for the window based on its position, but we haven't
-   * placed it yet.
+
+  /* FIXME should use xinerama with mouse pointer
+   * (or better, xinerama where window was launched
+   * determined via startup notification)
    */
-  meta_window_get_work_area (window, TRUE, &work_area);
-  
+  meta_window_get_work_area_for_xinerama (window, 0, &work_area);
+
   cascade_x = MAX (0, work_area.x);
   cascade_y = MAX (0, work_area.y);
-
-  /* Find first cascade position that's not used. */
-
-  /* arbitrary-ish threshold, honors user attempts to
-   * manually cascade.
-   */
-  if (fgeom)
-    {
-      x_threshold = MAX (fgeom->left_width, 15);
-      y_threshold = MAX (fgeom->top_height, 15);
-    }
-  else
-    {
-      x_threshold = 15;
-      y_threshold = 15;
-    }
   
+  /* Find first cascade position that's not used. */
+  
+  window_width = window->frame ? window->frame->rect.width : window->rect.width;
+  window_height = window->frame ? window->frame->rect.height : window->rect.height;
+  
+  cascade_stage = 0;
   tmp = sorted;
   while (tmp != NULL)
     {
@@ -147,7 +157,7 @@ find_next_cascade (MetaWindow *window,
           wx = w->rect.x;
           wy = w->rect.y;
         }
-
+      
       if (ABS (wx - cascade_x) < x_threshold &&
           ABS (wy - cascade_y) < y_threshold)
         {
@@ -158,6 +168,36 @@ find_next_cascade (MetaWindow *window,
           meta_window_get_position (w, &wx, &wy);
           cascade_x = wx;
           cascade_y = wy;
+          
+          /* If we go off the screen, start over with a new cascade */
+	  if (((cascade_x + window_width) >
+               (work_area.x + work_area.width)) ||
+              ((cascade_y + window_height) >
+	       (work_area.y + work_area.height)))
+	    {
+	      cascade_x = MAX (0, work_area.x);
+	      cascade_y = MAX (0, work_area.y);
+              
+#define CASCADE_INTERVAL 50 /* space between top-left corners of cascades */
+              cascade_stage += 1;
+	      cascade_x += CASCADE_INTERVAL * cascade_stage;
+              
+	      /* start over with a new cascade translated to the right, unless
+               * we are out of space
+               */
+              if ((cascade_x + window_width) <
+                  (work_area.x + work_area.width))
+                {
+                  tmp = sorted;
+                  continue;
+                }
+              else
+                {
+                  /* All out of space, this cascade_x won't work */
+                  cascade_x = MAX (0, work_area.x);
+                  break;
+                }
+	    }
         }
       else
         {
@@ -173,20 +213,6 @@ find_next_cascade (MetaWindow *window,
   
   g_list_free (sorted);
 
-  /* don't place windows off the screen */
-  if (((cascade_x + (window->frame ? window->frame->rect.width : window->rect.width)) >
-       (work_area.x + work_area.width)) ||
-      ((cascade_y + (window->frame ? window->frame->rect.height : window->rect.height)) >
-       (work_area.y + work_area.height)))
-    {
-      /* FIXME it would be better to try to start a new cascade to the
-       * right of the last one, probably. Instead of just falling back
-       * to "origin"
-       */
-      cascade_x = MAX (0, work_area.x);
-      cascade_y = MAX (0, work_area.y);
-    }
-  
   /* Convert coords to position of window, not position of frame. */
   if (fgeom == NULL)
     {
@@ -339,14 +365,36 @@ topmost_cmp (gconstpointer a, gconstpointer b)
     return 0;
 }
 
+static void
+center_tile_rect_in_area (MetaRectangle *rect,
+                          MetaRectangle *work_area)
+{
+  int fluff;
+
+  /* The point here is to tile a window such that "extra"
+   * space is equal on either side (i.e. so a full screen
+   * of windows tiled this way would center the windows
+   * as a group)
+   */
+  
+  fluff = (work_area->width % rect->width) / 2;
+  rect->x = work_area->x + fluff;
+  fluff = (work_area->height % rect->height) / 3;
+  rect->y = work_area->y + fluff;
+}
+
 static gboolean
-fit_rect_in_xinerama (MetaScreen    *screen,
+fit_rect_in_xinerama (MetaWindow    *window,
                       MetaRectangle *rect)
 {
   int i;
   int best_index;
   int best_overlap;
+  MetaScreen *screen;
+  MetaRectangle work_area;
   const MetaXineramaScreenInfo *xsi;
+
+  screen = window->screen;
   
   /* Find xinerama with best fit, then
    * shift rect to be entirely within it.
@@ -374,7 +422,10 @@ fit_rect_in_xinerama (MetaScreen    *screen,
         overlap = 0;
 
       if (overlap > best_overlap)
-        best_index = i; 
+        {
+          best_index = i;
+          best_overlap = overlap;
+        }
       
       ++i;
     }
@@ -382,17 +433,17 @@ fit_rect_in_xinerama (MetaScreen    *screen,
   /* some overlap had to be better than -1 */
   g_assert (best_index >= 0);
 
-  xsi = &screen->xinerama_infos[best_index];
+  meta_window_get_work_area_for_xinerama (window, best_index, &work_area);
   
-  if (rect->x < xsi->x_origin)
-    rect->x = xsi->x_origin;
-  if (rect->y < xsi->y_origin)
-    rect->y = xsi->y_origin;
+  if ((rect->x < work_area.x) || (rect->y < work_area.y) ||
+      (rect->x >= work_area.x + work_area.width) ||
+      (rect->y >= work_area.y + work_area.height))
+    center_tile_rect_in_area (rect, &work_area);
 
-  /* Now return whether we are entirely within the xinerama screen */
-  return
-    ((rect->x + rect->width) < (xsi->x_origin + xsi->width)) &&
-    ((rect->y + rect->height) < (xsi->y_origin + xsi->height));
+  /* Now return whether we are entirely within the work area */
+  return 
+    ((rect->x + rect->width) < (work_area.x + work_area.width)) &&
+    ((rect->y + rect->height) < (work_area.y + work_area.height));
 }
 
 /* Find the leftmost, then topmost, empty area on the workspace
@@ -419,15 +470,17 @@ find_first_fit (MetaWindow *window,
    * the bottom of each existing window, and then to the right
    * of each existing window, aligned with the left/top of the
    * existing window in each of those cases.
-   */
-  
+   */  
   int retval;
   GList *sorted;
   GList *tmp;
   MetaRectangle rect;
+  MetaRectangle work_area;
+  int i;
   
   retval = FALSE;
-
+  sorted = NULL;
+  
   rect.width = window->rect.width;
   rect.height = window->rect.height;
   
@@ -436,6 +489,31 @@ find_first_fit (MetaWindow *window,
       rect.width += fgeom->left_width + fgeom->right_width;
       rect.height += fgeom->top_height + fgeom->bottom_height;
     }
+
+  /* Try center-tiling on first xinerama */
+  /* FIXME should use xinerama with mouse pointer
+   * (or better, xinerama where window was launched
+   * determined via startup notification)
+   */
+  meta_window_get_work_area_for_xinerama (window, 0, &work_area);
+
+  center_tile_rect_in_area (&rect, &work_area);
+
+  if (fit_rect_in_xinerama (window, &rect) &&
+      !rectangle_overlaps_some_window (&rect, windows))
+    {
+      *new_x = rect.x;
+      *new_y = rect.y;
+      if (fgeom)
+        {
+          *new_x += fgeom->left_width;
+          *new_y += fgeom->top_height;
+        }
+      
+      retval = TRUE;
+      
+      goto out;
+    }
   
   sorted = g_list_copy (windows);
 
@@ -448,13 +526,13 @@ find_first_fit (MetaWindow *window,
     {
       MetaWindow *w = tmp->data;
       MetaRectangle outer_rect;
-      
-      meta_window_get_outer_rect (w, &outer_rect);
 
+      meta_window_get_outer_rect (w, &outer_rect);
+      
       rect.x = outer_rect.x;
       rect.y = outer_rect.y + outer_rect.height;
       
-      if (fit_rect_in_xinerama (window->screen, &rect) &&
+      if (fit_rect_in_xinerama (window, &rect) &&
           !rectangle_overlaps_some_window (&rect, sorted))
         {
           *new_x = rect.x;
@@ -464,12 +542,12 @@ find_first_fit (MetaWindow *window,
               *new_x += fgeom->left_width;
               *new_y += fgeom->top_height;
             }
-
+          
           retval = TRUE;
-
+          
           goto out;
-        }      
-      
+        }
+
       tmp = tmp->next;
     }
 
@@ -482,13 +560,13 @@ find_first_fit (MetaWindow *window,
     {
       MetaWindow *w = tmp->data;
       MetaRectangle outer_rect;
-      
+     
       meta_window_get_outer_rect (w, &outer_rect);
-
+     
       rect.x = outer_rect.x + outer_rect.width;
       rect.y = outer_rect.y;
-      
-      if (fit_rect_in_xinerama (window->screen, &rect) &&
+     
+      if (fit_rect_in_xinerama (window, &rect) &&
           !rectangle_overlaps_some_window (&rect, sorted))
         {
           *new_x = rect.x;
@@ -498,14 +576,41 @@ find_first_fit (MetaWindow *window,
               *new_x += fgeom->left_width;
               *new_y += fgeom->top_height;
             }
-
+         
           retval = TRUE;
-
+         
           goto out;
-        }      
-      
+        }
+
       tmp = tmp->next;
-    }  
+    }
+
+  /* Try center-tile on each Xinerama screen which isn't the first */
+  i = 1;
+  while (i < window->screen->n_xinerama_infos)
+    {
+      meta_window_get_work_area_for_xinerama (window, i, &work_area);
+
+      center_tile_rect_in_area (&rect, &work_area);
+
+      if (fit_rect_in_xinerama (window, &rect) &&
+          !rectangle_overlaps_some_window (&rect, windows))
+        {
+          *new_x = rect.x;
+          *new_y = rect.y;
+          if (fgeom)
+            {
+              *new_x += fgeom->left_width;
+              *new_y += fgeom->top_height;
+            }
+      
+          retval = TRUE;
+          
+          goto out;
+        }
+      
+      ++i;
+    }
   
  out:
   g_list_free (sorted);
@@ -526,22 +631,37 @@ constrain_placement (MetaWindow        *window,
    */
   MetaRectangle work_area;  
   int nw_x, nw_y;
+  int offscreen_w, offscreen_h;
+  MetaRectangle outer_rect;
+
+  meta_window_get_outer_rect (window, &outer_rect);
 
   /* FIXME this is bogus because we get the current xinerama
    * for the window based on its position, but we haven't
    * placed it yet.
    */
-  meta_window_get_work_area (window, TRUE, &work_area);
+  meta_window_get_work_area_current_xinerama (window, &work_area);
 
   nw_x = work_area.x;
   nw_y = work_area.y;
+
   if (window->frame)
     {
       nw_x += fgeom->left_width;
       nw_y += fgeom->top_height;
     }
 
-  /* Keep window from going off left edge, though we don't have
+  /* Keep window from going off the bottom right, though we don't have
+   * this constraint once the window has been placed
+   */
+  offscreen_w = (outer_rect.x + outer_rect.width) - (work_area.x + work_area.width);
+  if (offscreen_w > 0)
+    nw_x -= offscreen_w;
+  offscreen_h = (outer_rect.y + outer_rect.height) - (work_area.y + work_area.height);
+  if (offscreen_h > 0)
+    nw_y -= offscreen_h;
+  
+  /* Keep window from going off left edge, though again we don't have
    * this constraint once the window has been placed.
    */
   if (x < nw_x)
@@ -761,6 +881,7 @@ meta_window_place (MetaWindow        *window,
   constrain_placement (window, fgeom, x, y, &x, &y);
 
  done_no_constraints:
+
   *new_x = x;
   *new_y = y;
 }
@@ -858,13 +979,11 @@ get_vertical_edges (MetaWindow *window,
   edges = g_new (int, n_edges);
 
   /* workspace/screen edges */
-  meta_window_get_work_area (window, FALSE, &work_area);
+  meta_window_get_work_area_current_xinerama (window, &work_area);
 
   edges[i] = work_area.x;
   ++i;
-  edges[i] =
-    work_area.x +
-    work_area.width;
+  edges[i] = work_area.x + work_area.width;
   ++i;
   edges[i] = 0;
   ++i;
@@ -931,13 +1050,11 @@ get_horizontal_edges (MetaWindow *window,
   edges = g_new (int, n_edges);
 
   /* workspace/screen edges */
-  meta_window_get_work_area (window, FALSE, &work_area);
+  meta_window_get_work_area_current_xinerama (window, &work_area);
   
   edges[i] = work_area.y;
   ++i;
-  edges[i] =
-    work_area.y +
-    work_area.height;
+  edges[i] = work_area.y + work_area.height;
   ++i;
   edges[i] = 0;
   ++i;
