@@ -51,28 +51,6 @@ meta_ui_init (int *argc, char ***argv)
   if (!gtk_init_check (argc, argv))
     meta_fatal ("Unable to open X display %s\n", XDisplayName (NULL));
 
-  {
-    /* FIXME hackaround for Pango opening a separate display
-     * connection and doing a server grab while we have a grab
-     * on the primary display connection. This forces Pango to
-     * go ahead and do its font cache before we try to grab
-     * the server.
-     */  
-    PangoFontMetrics *metrics;
-    PangoLanguage *lang;
-    PangoContext *context;
-    PangoFontDescription *font_desc;
-
-    context = gdk_pango_context_get ();
-    lang = gtk_get_default_language ();
-    font_desc = pango_font_description_from_string ("Sans 12");
-    metrics = pango_context_get_metrics (context, font_desc, lang);    
-    
-    pango_font_metrics_unref (metrics);
-    pango_font_description_free (font_desc);
-    g_object_unref (G_OBJECT (context));
-  }
-
   meta_stock_icons_init ();
 }
 
@@ -238,12 +216,15 @@ meta_ui_reset_frame_bg (MetaUI *ui,
 }
 
 void
-meta_ui_apply_frame_shape  (MetaUI *ui,
-                            Window  xwindow,
-                            int     new_window_width,
-                            int     new_window_height)
+meta_ui_apply_frame_shape  (MetaUI  *ui,
+                            Window   xwindow,
+                            int      new_window_width,
+                            int      new_window_height,
+                            gboolean window_has_shape)
 {
-  meta_frames_apply_shapes (ui->frames, xwindow, new_window_width, new_window_height);
+  meta_frames_apply_shapes (ui->frames, xwindow,
+                            new_window_width, new_window_height,
+                            window_has_shape);
 }
 
 void
@@ -410,7 +391,11 @@ get_cmap (GdkPixmap *pixmap)
       else
         {
           meta_verbose ("Using system cmap to snapshot pixmap\n");
+#ifdef HAVE_GTK_MULTIHEAD
+          cmap = gdk_screen_get_system_colormap (gdk_drawable_get_screen (pixmap));
+#else
           cmap = gdk_colormap_get_system ();
+#endif
           g_object_ref (G_OBJECT (cmap));
         }
     }
