@@ -41,10 +41,10 @@ struct _MetaWorkspace
   
   GList *windows;
 
-  char *name;
+  MetaRectangle all_work_areas;
   
-  MetaRectangle work_area;
-  guint work_area_invalid : 1;
+  MetaRectangle *work_areas;
+  guint work_areas_invalid : 1;
 };
 
 MetaWorkspace* meta_workspace_new           (MetaScreen    *screen);
@@ -63,15 +63,22 @@ int            meta_workspace_index         (MetaWorkspace *workspace);
 GList*         meta_workspace_list_windows  (MetaWorkspace *workspace);
 
 void meta_workspace_invalidate_work_area (MetaWorkspace *workspace);
-void meta_workspace_get_work_area        (MetaWorkspace *workspace,
-                                          MetaRectangle *area);
+
+
+void meta_workspace_get_work_area_for_xinerama  (MetaWorkspace *workspace,
+                                                 int            which_xinerama,
+                                                 MetaRectangle *area);
+void meta_workspace_get_work_area_all_xineramas (MetaWorkspace *workspace,
+                                                 MetaRectangle *area);
+
+
+
 
 
 MetaWorkspace* meta_workspace_get_neighbor (MetaWorkspace      *workspace,
                                             MetaMotionDirection direction);
 
-void meta_workspace_set_name (MetaWorkspace *workspace,
-                              const char    *name);
+const char* meta_workspace_get_name (MetaWorkspace *workspace);
 
 #endif
 
